/*****************************************************************************
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: if-1.cpp,v 1.2 2000/06/07 03:50:27 kevin Exp $
**  $Log: if-1.cpp,v $
**  Revision 1.2  2000/06/07 03:50:27  kevin
**  *** empty log message ***
**
**  Revision 1.1  2000/06/07 02:29:05  kevin
**  Initial C++ versions
**
**  Revision 1.6  2000/05/24 22:50:04  kevin
**  Added support for new SGP library
**
**  Revision 1.5  2000/05/16 04:33:59  kevin
**  Improved option processing
**
**  Revision 1.4  2000/05/09 14:52:27  kevin
**  added sqr and sqrt functions
**
**  Revision 1.3  2000/05/08 20:02:32  kevin
**  ANSI C changes
**
**  Revision 1.2  2000/05/03 08:49:50  kevin
**  Code cleanup
**
**  Revision 1.1.1.1  2000/04/28 13:02:44  kevin
**  Initial CVS import for first public release
**
**
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

/* FILE
 *   if-1.c             Filter a single IF file
 */

#include "ct.h"

enum {O_LOG, O_EXP, O_SQRT, O_SQR, O_INVERT, O_VERBOSE, O_HELP, O_VERSION};

static struct option my_options[] =
{
  {"invert", 0, 0, O_INVERT},
  {"verbose", 0, 0, O_VERBOSE},
  {"log", 0, 0, O_LOG},
  {"exp", 0, 0, O_EXP},
  {"sqr", 0, 0, O_SQR},
  {"sqrt", 0, 0, O_SQRT},
  {"help", 0, 0, O_HELP},
  {"version", 0, 0, O_VERSION},
  {0, 0, 0, 0}
};

void 
if1_usage (const char *program)
{
  fprintf(stdout, "if1_usage: %s infile outfile [OPTIONS]\n", kbasename(program));
  fprintf(stdout, "Generate a IF file from a IF file\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "     --invert   Invert image\n");
  fprintf(stdout, "     --log      Natural logrithm of image\n");
  fprintf(stdout, "     --exp      Natural exponential of image\n");
  fprintf(stdout, "     --sqr      Square of image\n");
  fprintf(stdout, "     --sqrt     Square root of image\n");
  fprintf(stdout, "     --verbose  Verbose modem\n");
  fprintf(stdout, "     --version  Print version\n");
  fprintf(stdout, "     --help     Print this help message\n");
}

int 
if1_main (int argc, char *const argv[])
{
  ImageFile *im_in;
  ImageFile *im_out;
  char *in_file;
  char *out_file;
  int opt_verbose = 0;
  int opt_invert = 0;
  int opt_log = 0;
  int opt_exp = 0;
  int opt_sqr = 0;
  int opt_sqrt = 0;

  while (1)
    {
      int c = getopt_long (argc, argv, "", my_options, NULL);
      
      if (c == -1)
	break;
      
      switch (c)
	{
	case O_INVERT:
	  opt_invert = 1;
	  break;
	case O_LOG:
	  opt_log = 1;
	  break;
	case O_SQR:
	  opt_sqr = 1;
	  break;
	case O_SQRT:
	  opt_sqrt = 1;
	  break;
	case O_EXP:
	  opt_exp = 1;
	  break;
	case O_VERBOSE:
	  opt_verbose = 1;
	  break;
        case O_VERSION:
#ifdef VERSION
	  fprintf(stdout, "Version %s\n", VERSION);
#else
	  fprintf(stderr, "Unknown version number");
#endif
	  exit(0);
	case O_HELP:
	case '?':
	  if1_usage(argv[0]);
	  return (0);
	default:
	  if1_usage(argv[0]);
	  return (1);
	}
    }

  if (optind + 2 != argc)
    {
      if1_usage(argv[0]);
      return (1);
    }
  
  in_file = argv[optind];
  out_file = argv[optind + 1];


  string histString;

  if (opt_invert || opt_log || opt_exp || opt_sqr || opt_sqrt) {
    int ix, iy;

    im_in = new ImageFile (in_file);
    im_in->adf.fileRead ();
    int nx = im_in->adf.nx();
    int ny = im_in->adf.ny();
    im_out = new ImageFile (out_file, nx, ny);
    im_out->adf.fileCreate ();

    ImageFileArray vIn = im_in->adf.getArray();
    ImageFileArray vOut = im_out->adf.getArray();

    if (opt_invert) {
      for (ix = 0; ix < nx; ix++)
        for (iy = 0; iy < ny; iy++)
          vOut[ix][iy] = - vIn[ix][iy];
      histString = "Invert transformation";
    }
    if (opt_log) {
      for (ix = 0; ix < nx; ix++)
        for (iy = 0; iy < ny; iy++)
          vOut[ix][iy] = log (vIn[ix][iy]);
      histString = "Log transformation";
    }
    if (opt_exp) {
      for (ix = 0; ix < nx; ix++)
        for (iy = 0; iy < ny; iy++)
          vOut[ix][iy] = exp (vIn[ix][iy]);
      histString = "Exp transformation";
    }
    if (opt_sqr) {
      for (ix = 0; ix < nx; ix++)
        for (iy = 0; iy < ny; iy++)
          vOut[ix][iy] = vIn[ix][iy] * vIn[ix][iy];
      histString = "Sqr transformation";
    }
    if (opt_sqrt) {
      for (ix = 0; ix < nx; ix++)
        for (iy = 0; iy < ny; iy++)
          vOut[ix][iy] = sqrt (vIn[ix][iy]);
      histString = "Sqrt transformation";
    }

    im_out->adf.arrayDataWrite ();
    for (int i = 0; i < im_in->adf.getNumLabels(); i++) {
      Array2dFileLabel l;
      im_in->adf.labelRead (l, i);
      im_out->adf.labelAdd (l);
    }
    im_out->adf.labelAdd (Array2dFileLabel::L_HISTORY, histString.c_str());
    im_out->adf.fileClose ();
  }

  return (0);
}

#ifndef NO_MAIN
int 
main (int argc, char *const argv[])
{
  return (if1_main(argc, argv));
}
#endif

